/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var SankeyDiagram1446463184954;
        (function (SankeyDiagram1446463184954) {
            var SelectionManager = visuals.utility.SelectionManager;
            var ValueFormatter = powerbi.visuals.valueFormatter;
            var SankeyDiagram = (function () {
                function SankeyDiagram(constructorOptions) {
                    this.margin = {
                        top: 10,
                        right: 10,
                        bottom: 10,
                        left: 10
                    };
                    this.nodeWidth = 21.5;
                    this.curvatureOfLinks = 0.5;
                    if (constructorOptions) {
                        this.svg = constructorOptions.svg;
                        this.margin = constructorOptions.margin || this.margin;
                        this.curvatureOfLinks = constructorOptions.curvatureOfLinks || this.curvatureOfLinks;
                    }
                }
                SankeyDiagram.getProperties = function (capabilities) {
                    var result = {};
                    for (var objectKey in capabilities.objects) {
                        result[objectKey] = {};
                        for (var propKey in capabilities.objects[objectKey].properties) {
                            result[objectKey][propKey] = {
                                objectName: objectKey,
                                propertyName: propKey
                            };
                        }
                    }
                    return result;
                };
                Object.defineProperty(SankeyDiagram.prototype, "textProperties", {
                    get: function () {
                        return {
                            fontFamily: this.root.style("font-family"),
                            fontSize: jsCommon.PixelConverter.fromPoint(this.dataView ? this.dataView.settings.fontSize : SankeyDiagram.DefaultSettings.fontSize)
                        };
                    },
                    enumerable: true,
                    configurable: true
                });
                SankeyDiagram.prototype.init = function (visualsInitOptions) {
                    if (this.svg) {
                        this.root = this.svg;
                    }
                    else {
                        this.root = d3.select(visualsInitOptions.element.get(0)).append("svg");
                    }
                    this.selectionManager = new SelectionManager({ hostServices: visualsInitOptions.host });
                    var style = visualsInitOptions.style;
                    this.colours = style && style.colorPalette ? style.colorPalette.dataColors : new visuals.DataColorPalette();
                    this.root.classed(SankeyDiagram.ClassName, true);
                    this.main = this.root.append("g");
                    this.links = this.main.append("g").classed(SankeyDiagram.Links["class"], true);
                    this.nodes = this.main.append("g").classed(SankeyDiagram.Nodes["class"], true);
                };
                SankeyDiagram.prototype.update = function (visualUpdateOptions) {
                    if (!visualUpdateOptions || !visualUpdateOptions.dataViews) {
                        return;
                    }
                    var dataView = visualUpdateOptions.dataViews[0], sankeyDiagramDataView;
                    this.updateViewport(visualUpdateOptions.viewport);
                    sankeyDiagramDataView = this.converter(dataView);
                    this.findNodePosition(sankeyDiagramDataView);
                    this.dataView = sankeyDiagramDataView;
                    this.render(sankeyDiagramDataView);
                };
                SankeyDiagram.prototype.updateViewport = function (viewport) {
                    var height, width;
                    height = this.getPositiveNumber(viewport.height);
                    width = this.getPositiveNumber(viewport.width);
                    this.viewport = {
                        height: this.getPositiveNumber(height - this.margin.top - this.margin.bottom),
                        width: this.getPositiveNumber(width - this.margin.left - this.margin.right)
                    };
                    this.updateElements(height, width);
                };
                SankeyDiagram.prototype.getPositiveNumber = function (value) {
                    return value < 0 || isNaN(value) || value === Infinity || value === -Infinity ? 0 : value;
                };
                SankeyDiagram.prototype.updateElements = function (height, width) {
                    this.root.attr({
                        "height": height,
                        "width": width
                    });
                    this.main.attr("transform", visuals.SVGUtil.translate(this.margin.left, this.margin.top));
                };
                SankeyDiagram.prototype.converter = function (dataView) {
                    var _this = this;
                    if (!dataView || !dataView.categorical || !dataView.categorical.categories || !dataView.categorical.categories[0] || !dataView.categorical.categories[1] || !dataView.categorical.categories[0].values || !dataView.categorical.categories[1].values) {
                        return {
                            nodes: [],
                            links: [],
                            settings: {
                                scale: { x: 1, y: 1 },
                                colourOfLabels: SankeyDiagram.DefaultSettings.colourOfLabels,
                                fontSize: SankeyDiagram.DefaultSettings.fontSize
                            }
                        };
                    }
                    var nodes = [], links = [], dataPoints = [], categories = dataView.categorical.categories[0].values, secondCategories = dataView.categorical.categories[1].values, valuesColumn = dataView.categorical.values && dataView.categorical.values[0], weightValues = valuesColumn && valuesColumn.values && valuesColumn.values.map(function (x) { return x || 0; }) || [], allCategories, valueFormatterForCategories, formatOfWeigth = "g", valuesFormatterForWeigth, objects, linksObjects = dataView.categorical.categories[0].objects || [], labelColour, settings, shiftOfColour, identities = [];
                    if (dataView.categorical.categories[0].identity) {
                        identities = identities.concat(dataView.categorical.categories[0].identity);
                    }
                    if (dataView.categorical.categories[1].identity) {
                        identities = identities.concat(dataView.categorical.categories[1].identity);
                    }
                    objects = this.getObjectsFromDataView(dataView);
                    labelColour = this.getColour(SankeyDiagram.Properties["labels"]["fill"], SankeyDiagram.DefaultSettings.colourOfLabels, objects);
                    if (valuesColumn && valuesColumn.source) {
                        formatOfWeigth = ValueFormatter.getFormatString(valuesColumn.source, SankeyDiagram.Properties["general"]["formatString"]);
                    }
                    dataPoints = categories.map(function (item, index) {
                        return {
                            source: item,
                            destination: secondCategories[index],
                            weigth: valuesColumn ? Math.max(weightValues[index] || 0, 0) : 1
                        };
                    });
                    allCategories = categories.concat(secondCategories);
                    valueFormatterForCategories = ValueFormatter.create({
                        format: ValueFormatter.getFormatString(dataView.categorical.categories[0].source, SankeyDiagram.Properties["general"]["formatString"]),
                        value: allCategories[0],
                        value2: allCategories[allCategories.length - 1]
                    });
                    valuesFormatterForWeigth = ValueFormatter.create({
                        format: formatOfWeigth,
                        value: Math.max(d3.max(weightValues) || 1, 1),
                    });
                    allCategories.forEach(function (item, index) {
                        if (!nodes.some(function (node) {
                            if (item === node.label.name) {
                                node.selectionIds.push(visuals.SelectionId.createWithId(identities[index]));
                                return true;
                            }
                            return false;
                        })) {
                            var formattedValue = valueFormatterForCategories.format(item), label, selectionId, textProperties = {
                                text: formattedValue,
                                fontFamily: _this.textProperties.fontFamily,
                                fontSize: _this.textProperties.fontSize
                            };
                            label = {
                                name: item,
                                formattedName: valueFormatterForCategories.format(item),
                                width: powerbi.TextMeasurementService.measureSvgTextWidth(textProperties),
                                height: powerbi.TextMeasurementService.estimateSvgTextHeight(textProperties),
                                colour: labelColour
                            };
                            selectionId = visuals.SelectionId.createWithId(identities[index]);
                            nodes.push({
                                label: label,
                                links: [],
                                inputWeight: 0,
                                outputWeight: 0,
                                width: _this.nodeWidth,
                                height: 0,
                                colour: SankeyDiagram.DefaultColourOfNode,
                                tooltipData: [],
                                selectionIds: [selectionId]
                            });
                        }
                    });
                    shiftOfColour = this.colours.getAllColors().length / nodes.length;
                    nodes.forEach(function (node, index) {
                        node.colour = _this.colours.getColorByIndex(Math.floor(index * shiftOfColour)).value;
                    });
                    dataPoints.forEach(function (dataPoint, index) {
                        var sourceNode, destinationNode, link, linkColour;
                        if (dataPoint.source === dataPoint.destination) {
                            return;
                        }
                        nodes.forEach(function (node) {
                            if (node.label.name === dataPoint.source) {
                                sourceNode = node;
                            }
                            if (node.label.name === dataPoint.destination) {
                                destinationNode = node;
                            }
                        });
                        linkColour = _this.getColour(SankeyDiagram.Properties["links"]["fill"], SankeyDiagram.DefaultColourOfLink, linksObjects[index]);
                        link = {
                            source: sourceNode,
                            destination: destinationNode,
                            weigth: dataPoint.weigth,
                            height: dataPoint.weigth,
                            colour: linkColour,
                            tooltipData: _this.getTooltipDataForLink(valuesFormatterForWeigth, sourceNode.label.formattedName, destinationNode.label.formattedName, dataPoint.weigth),
                            selectionId: visuals.SelectionId.createWithId(identities[index])
                        };
                        links.push(link);
                        sourceNode.links.push(link);
                        destinationNode.links.push(link);
                        _this.updateValueOfNode(sourceNode);
                        _this.updateValueOfNode(destinationNode);
                        sourceNode.tooltipData = _this.getTooltipForNode(valuesFormatterForWeigth, sourceNode.label.formattedName, sourceNode.inputWeight ? sourceNode.inputWeight : sourceNode.outputWeight);
                        destinationNode.tooltipData = _this.getTooltipForNode(valuesFormatterForWeigth, destinationNode.label.formattedName, destinationNode.inputWeight ? destinationNode.inputWeight : destinationNode.outputWeight);
                    });
                    settings = this.parseSettings(objects);
                    settings.colourOfLabels = labelColour;
                    return {
                        nodes: nodes,
                        links: links,
                        settings: settings
                    };
                };
                SankeyDiagram.prototype.getObjectsFromDataView = function (dataView) {
                    if (!dataView || !dataView.metadata || !dataView.metadata.columns || !dataView.metadata.objects) {
                        return null;
                    }
                    return dataView.metadata.objects;
                };
                SankeyDiagram.prototype.getColour = function (properties, defaultColor, objects) {
                    var colorHelper;
                    colorHelper = new visuals.ColorHelper(this.colours, properties, defaultColor);
                    return colorHelper.getColorForMeasure(objects, "");
                };
                SankeyDiagram.prototype.getTooltipDataForLink = function (valueFormatter, sourceNodeName, destinationNodeName, linkWeight) {
                    var formattedLinkWeight;
                    if (valueFormatter && valueFormatter.format) {
                        formattedLinkWeight = valueFormatter.format(linkWeight);
                    }
                    else {
                        formattedLinkWeight = linkWeight.toString();
                    }
                    return [{
                        displayName: SankeyDiagram.RoleNames.rows,
                        value: sourceNodeName
                    }, {
                        displayName: SankeyDiagram.RoleNames.columns,
                        value: destinationNodeName
                    }, {
                        displayName: SankeyDiagram.RoleNames.values,
                        value: formattedLinkWeight
                    }];
                };
                SankeyDiagram.prototype.updateValueOfNode = function (node) {
                    node.inputWeight = node.links.reduce(function (previousValue, currentValue) {
                        return previousValue + (currentValue.destination === node ? currentValue.weigth : 0);
                    }, 0);
                    node.outputWeight = node.links.reduce(function (previousValue, currentValue) {
                        return previousValue + (currentValue.source === node ? currentValue.weigth : 0);
                    }, 0);
                };
                SankeyDiagram.prototype.getTooltipForNode = function (valueFormatter, nodeName, nodeWeight) {
                    var formattedNodeWeigth;
                    if (valueFormatter && valueFormatter.format) {
                        formattedNodeWeigth = valueFormatter.format(nodeWeight);
                    }
                    else {
                        formattedNodeWeigth = nodeWeight.toString();
                    }
                    return [{
                        displayName: "Name",
                        value: nodeName
                    }, {
                        displayName: SankeyDiagram.RoleNames.values,
                        value: formattedNodeWeigth
                    }];
                };
                SankeyDiagram.prototype.parseSettings = function (objects) {
                    var isVisibleLabels = false;
                    isVisibleLabels = powerbi.DataViewObjects.getValue(objects, SankeyDiagram.Properties["labels"]["show"], SankeyDiagram.DefaultSettings.isVisibleLabels);
                    return {
                        isVisibleLabels: isVisibleLabels,
                        scale: {
                            x: SankeyDiagram.DefaultSettings.scale.x,
                            y: SankeyDiagram.DefaultSettings.scale.y
                        },
                        colourOfLabels: SankeyDiagram.DefaultSettings.colourOfLabels,
                        fontSize: powerbi.DataViewObjects.getValue(objects, SankeyDiagram.Properties["labels"]["fontSize"], SankeyDiagram.DefaultSettings.fontSize)
                    };
                };
                SankeyDiagram.prototype.findNodePosition = function (sankeyDiagramDataView) {
                    this.findNodePositionByX(sankeyDiagramDataView);
                    this.findNodePositionByY(sankeyDiagramDataView);
                };
                SankeyDiagram.prototype.findNodePositionByX = function (sankeyDiagramDataView) {
                    var nodes = sankeyDiagramDataView.nodes, nextNodes = [], previousNodes = [], x = 0, isRecursiveDependencies = false;
                    while (nodes.length > 0) {
                        nextNodes = [];
                        nodes.forEach(function (node) {
                            node.x = x;
                            node.links.forEach(function (link) {
                                if (node === link.source && node !== link.destination) {
                                    if (nextNodes.every(function (item) {
                                        return item !== link.destination;
                                    })) {
                                        nextNodes.push(link.destination);
                                    }
                                }
                            });
                        });
                        isRecursiveDependencies = nextNodes.length === previousNodes.length && previousNodes.every(function (previousNode) {
                            return nextNodes.some(function (nextNode) {
                                return nextNode === previousNode;
                            });
                        });
                        if (isRecursiveDependencies) {
                            previousNodes.forEach(function (element) {
                                element.x = x;
                                x++;
                            });
                            nodes = [];
                        }
                        else {
                            nodes = nextNodes;
                            previousNodes = nodes;
                            x++;
                        }
                    }
                    sankeyDiagramDataView.settings.scale.x = this.getScaleByAxisX(x - 1);
                    this.scaleByAxisX(sankeyDiagramDataView.nodes, sankeyDiagramDataView.settings.scale.x);
                };
                SankeyDiagram.prototype.scaleByAxisX = function (nodes, scale) {
                    nodes.forEach(function (node) {
                        node.x *= scale;
                    });
                };
                SankeyDiagram.prototype.getScaleByAxisX = function (numberOfColumns) {
                    if (numberOfColumns === void 0) { numberOfColumns = 1; }
                    return this.getPositiveNumber((this.viewport.width - this.nodeWidth) / numberOfColumns);
                };
                SankeyDiagram.prototype.findNodePositionByY = function (sankeyDiagramDataView) {
                    var nodes = sankeyDiagramDataView.nodes, links = sankeyDiagramDataView.links, currentX = 0, index = 0, maxIndex = 0, sumValueOfNodes = 0, maxValueOfNodes = 0;
                    nodes = nodes.sort(function (firstNode, secondNode) {
                        return firstNode.x - secondNode.x;
                    });
                    nodes.forEach(function (node) {
                        if (currentX !== node.x) {
                            index = 0;
                            currentX = node.x;
                            sumValueOfNodes = 0;
                        }
                        sumValueOfNodes += Math.max(node.inputWeight, node.outputWeight);
                        if (sumValueOfNodes > maxValueOfNodes) {
                            maxValueOfNodes = sumValueOfNodes;
                        }
                        if (index > maxIndex) {
                            maxIndex = index;
                        }
                        index++;
                    });
                    sankeyDiagramDataView.settings.scale.y = this.getScaleByAxisY(maxIndex + 1, maxValueOfNodes);
                    this.scaleByAxisY(nodes, links, sankeyDiagramDataView.settings.scale.y);
                };
                SankeyDiagram.prototype.getScaleByAxisY = function (numberOfRows, sumValueOfNodes) {
                    return this.getPositiveNumber((this.viewport.height - numberOfRows * SankeyDiagram.NodePadding) / sumValueOfNodes);
                };
                SankeyDiagram.prototype.scaleByAxisY = function (nodes, links, scale) {
                    var shiftByAxisY = 0, currentX = 0, index = 0;
                    nodes.forEach(function (node) {
                        if (currentX !== node.x) {
                            currentX = node.x;
                            shiftByAxisY = 0;
                            index = 0;
                        }
                        node.height = Math.max(node.inputWeight, node.outputWeight) * scale;
                        node.y = shiftByAxisY + SankeyDiagram.NodePadding * index;
                        shiftByAxisY += node.height;
                        index++;
                    });
                    nodes.forEach(function (node) {
                        node.links = node.links.sort(function (firstLink, secondLink) {
                            var firstY, secondY;
                            firstY = firstLink.source === node ? firstLink.destination.y : firstLink.source.y;
                            secondY = secondLink.source === node ? secondLink.destination.y : secondLink.source.y;
                            return firstY - secondY;
                        });
                        var shiftByAxisYOfLeftLink = 0, shiftByAxisYOfRightLink = 0;
                        node.links.forEach(function (link) {
                            var shiftByAxisY = 0;
                            link.height = link.weigth * scale;
                            if (link.source.x < node.x || link.destination.x < node.x) {
                                shiftByAxisY = shiftByAxisYOfLeftLink;
                                shiftByAxisYOfLeftLink += link.height;
                            }
                            else if (link.source.x > node.x || link.destination.x > node.x) {
                                shiftByAxisY = shiftByAxisYOfRightLink;
                                shiftByAxisYOfRightLink += link.height;
                            }
                            if (link.source === node) {
                                link.dySource = shiftByAxisY;
                            }
                            else if (link.destination === node) {
                                link.dyDestination = shiftByAxisY;
                            }
                        });
                    });
                };
                SankeyDiagram.prototype.render = function (sankeyDiagramDataView) {
                    var nodesSelection, linksSelection;
                    linksSelection = this.renderLinks(sankeyDiagramDataView);
                    nodesSelection = this.renderNodes(sankeyDiagramDataView);
                    this.bindSelectionHandler(sankeyDiagramDataView, nodesSelection, linksSelection);
                };
                SankeyDiagram.prototype.renderNodes = function (sankeyDiagramDataView) {
                    var _this = this;
                    var nodesEnterSelection, nodesSelection, nodeElements;
                    nodeElements = this.main.select(SankeyDiagram.Nodes.selector).selectAll(SankeyDiagram.Node.selector);
                    nodesSelection = nodeElements.data(sankeyDiagramDataView.nodes.filter(function (x) { return x.height > 0; }));
                    nodesEnterSelection = nodesSelection.enter().append("g");
                    nodesSelection.attr("transform", function (node) {
                        return visuals.SVGUtil.translate(node.x, node.y);
                    }).classed(SankeyDiagram.Node["class"], true);
                    nodesEnterSelection.append("rect").classed(SankeyDiagram.NodeRect["class"], true);
                    nodesEnterSelection.append("text").classed(SankeyDiagram.NodeLabel["class"], true);
                    nodesSelection.select(SankeyDiagram.NodeRect.selector).style({
                        "fill": function (node) { return node.colour; },
                        "stroke": function (node) { return d3.rgb(node.colour).darker(1.5); }
                    }).attr({
                        x: 0,
                        y: 0,
                        height: function (node) { return node.height; },
                        width: function (node) { return node.width; }
                    });
                    nodesSelection.select(SankeyDiagram.NodeLabel.selector).attr({
                        x: function (node) { return _this.getLabelPositionByAxisX(node); },
                        y: function (node) { return node.height / 2; },
                        dy: "0.35em"
                    }).style("fill", function (node) { return node.label.colour; }).style("font-size", this.textProperties.fontSize).style("display", function (node) {
                        var isNotVisibleLabel = false, labelPositionByAxisX = _this.getCurrentPositionOfLabelByAxisX(node);
                        isNotVisibleLabel = labelPositionByAxisX >= _this.viewport.width || labelPositionByAxisX <= 0 || (node.height + SankeyDiagram.NodePadding) < node.label.height;
                        if (isNotVisibleLabel || !sankeyDiagramDataView.settings.isVisibleLabels || sankeyDiagramDataView.settings.scale.x / 2 < SankeyDiagram.MinWidthOfLabel) {
                            return "none";
                        }
                        return null;
                    }).style("text-anchor", function (node) {
                        if (_this.isLabelLargerWidth(node)) {
                            return "end";
                        }
                        return null;
                    }).text(function (node) {
                        var maxWidth = sankeyDiagramDataView.settings.scale.x / 2 - node.width - SankeyDiagram.NodePadding;
                        if (_this.getCurrentPositionOfLabelByAxisX(node) > maxWidth) {
                            return powerbi.TextMeasurementService.getTailoredTextOrDefault({
                                text: node.label.formattedName,
                                fontFamily: _this.textProperties.fontFamily,
                                fontSize: _this.textProperties.fontSize
                            }, maxWidth);
                        }
                        return node.label.formattedName;
                    });
                    nodesSelection.exit().remove();
                    this.renderTooltip(nodesSelection);
                    return nodesSelection;
                };
                SankeyDiagram.prototype.getLabelPositionByAxisX = function (node) {
                    if (this.isLabelLargerWidth(node)) {
                        return -(SankeyDiagram.LabelPadding);
                    }
                    return node.width + SankeyDiagram.LabelPadding;
                };
                SankeyDiagram.prototype.isLabelLargerWidth = function (node) {
                    var shiftByAxisX = node.x + node.width + SankeyDiagram.LabelPadding;
                    return shiftByAxisX + node.label.width > this.viewport.width;
                };
                SankeyDiagram.prototype.getCurrentPositionOfLabelByAxisX = function (node) {
                    var labelPositionByAxisX = this.getLabelPositionByAxisX(node);
                    labelPositionByAxisX = labelPositionByAxisX > 0 ? labelPositionByAxisX + node.x + node.label.width + node.width : node.x + labelPositionByAxisX - node.label.width - node.width;
                    return labelPositionByAxisX;
                };
                SankeyDiagram.prototype.renderLinks = function (sankeyDiagramDataView) {
                    var _this = this;
                    var linksSelection, linksElements;
                    linksElements = this.main.select(SankeyDiagram.Links.selector).selectAll(SankeyDiagram.Link.selector);
                    linksSelection = linksElements.data(sankeyDiagramDataView.links.filter(function (x) { return x.height > 0; }));
                    linksSelection.enter().append("path").classed(SankeyDiagram.Link["class"], true);
                    linksSelection.attr("d", function (link) {
                        return _this.getSvgPath(link);
                    }).style({
                        "stroke-width": function (link) { return link.height; },
                        "stroke": function (link) { return link.colour; }
                    });
                    linksSelection.exit().remove();
                    this.renderTooltip(linksSelection);
                    return linksSelection;
                };
                SankeyDiagram.prototype.getSvgPath = function (link) {
                    var x0, x1, xi, x2, x3, y0, y1;
                    if (link.destination.x < link.source.x) {
                        x0 = link.source.x;
                        x1 = link.destination.x + link.destination.width;
                    }
                    else {
                        x0 = link.source.x + link.source.width;
                        x1 = link.destination.x;
                    }
                    xi = d3.interpolateNumber(x0, x1);
                    x2 = xi(this.curvatureOfLinks);
                    x3 = xi(1 - this.curvatureOfLinks);
                    y0 = link.source.y + link.dySource + link.height / 2;
                    y1 = link.destination.y + link.dyDestination + link.height / 2;
                    return "M " + x0 + " " + y0 + " C " + x2 + " " + y0 + ", " + x3 + " " + y1 + ", " + x1 + " " + y1;
                };
                SankeyDiagram.prototype.renderTooltip = function (selection) {
                    visuals.TooltipManager.addTooltip(selection, function (tooltipEvent) {
                        return tooltipEvent.data.tooltipData;
                    });
                };
                SankeyDiagram.prototype.bindSelectionHandler = function (sankeyDiagramDataView, nodesSelection, linksSelection) {
                    var _this = this;
                    nodesSelection.on("click", function (node) {
                        var selectionIds = _.isEqual(node.selectionIds.map(function (x) { return x.getKey(); }), _this.selectionManager.getSelectionIds().map(function (x) { return x.getKey(); })) ? [] : node.selectionIds;
                        _this.selectMany(selectionIds, true).then(function () { return _this.setSelection(nodesSelection, linksSelection); });
                        d3.event.stopPropagation();
                    });
                    linksSelection.on("click", function (link) {
                        _this.selectionManager.select(link.selectionId, d3.event.ctrlKey).then(function () { return _this.setSelection(nodesSelection, linksSelection); });
                        d3.event.stopPropagation();
                    });
                    this.root.on("click", function () {
                        _this.selectionManager.clear().then(function () { return _this.setSelection(nodesSelection, linksSelection); });
                    });
                    this.setSelection(nodesSelection, linksSelection);
                };
                SankeyDiagram.prototype.selectMany = function (selectionIds, clear) {
                    var _this = this;
                    if (clear === void 0) { clear = false; }
                    if (clear) {
                        return this.selectionManager.clear().then(function () { return _this.selectMany(selectionIds, false); });
                    }
                    var selectionDeffered = selectionIds.map(function (id) { return _this.selectionManager.select(id, true); });
                    return $.when.apply(null, selectionDeffered);
                };
                SankeyDiagram.prototype.setSelection = function (nodes, links) {
                    var selectionIds = this.selectionManager.getSelectionIds();
                    nodes.classed("selected", selectionIds.length > 0);
                    links.classed("selected", false);
                    if (selectionIds.length === 0) {
                        return;
                    }
                    var selectedNodesSelection = nodes.filter(function (node) { return selectionIds.some(function (selectedId) { return node.selectionIds.some(function (x) { return x.getKey() === selectedId.getKey(); }); }); });
                    var selectedLinksSelection = links.filter(function (link) { return selectionIds.some(function (selectionId) { return selectionId.getKey() === link.selectionId.getKey(); }); });
                    selectedNodesSelection.classed("selected", false);
                    selectedLinksSelection.classed("selected", true);
                };
                SankeyDiagram.prototype.enumerateObjectInstances = function (options) {
                    var enumeration = new visuals.ObjectEnumerationBuilder();
                    if (!this.dataView) {
                        return [];
                    }
                    switch (options.objectName) {
                        case "labels": {
                            this.enumerateLabels(enumeration);
                            break;
                        }
                        case "links": {
                            this.enumerateLinks(enumeration);
                            break;
                        }
                    }
                    return enumeration.complete();
                };
                SankeyDiagram.prototype.enumerateLabels = function (enumeration) {
                    var settings = this.dataView.settings, labels;
                    if (!settings) {
                        return;
                    }
                    labels = {
                        objectName: "labels",
                        displayName: "labels",
                        selector: null,
                        properties: {
                            show: settings.isVisibleLabels,
                            fill: settings.colourOfLabels,
                            fontSize: settings.fontSize
                        }
                    };
                    enumeration.pushInstance(labels);
                };
                SankeyDiagram.prototype.enumerateLinks = function (enumeration) {
                    var links = this.dataView.links;
                    if (!links || !(links.length > 0)) {
                        return;
                    }
                    links.forEach(function (link) {
                        enumeration.pushInstance({
                            objectName: "links",
                            displayName: "" + link.source.label.formattedName + " - " + link.destination.label.formattedName,
                            selector: visuals.ColorHelper.normalizeSelector(link.selectionId.getSelector(), false),
                            properties: {
                                fill: { solid: { color: link.colour } }
                            }
                        });
                    });
                };
                SankeyDiagram.ClassName = "sankeyDiagram";
                SankeyDiagram.Nodes = {
                    "class": "nodes",
                    selector: ".nodes"
                };
                SankeyDiagram.Node = {
                    "class": "node",
                    selector: ".node"
                };
                SankeyDiagram.NodeRect = {
                    "class": "nodeRect",
                    selector: ".nodeRect"
                };
                SankeyDiagram.NodeLabel = {
                    "class": "nodeLabel",
                    selector: ".nodeLabel"
                };
                SankeyDiagram.Links = {
                    "class": "links",
                    selector: ".links"
                };
                SankeyDiagram.Link = {
                    "class": "link",
                    selector: ".link"
                };
                SankeyDiagram.DefaultColourOfNode = "rgb(62, 187, 162)";
                SankeyDiagram.DefaultColourOfLink = "black";
                SankeyDiagram.DefaultSettings = {
                    isVisibleLabels: true,
                    scale: { x: 1, y: 1 },
                    colourOfLabels: "black",
                    fontSize: 12
                };
                SankeyDiagram.MinWidthOfLabel = 35;
                SankeyDiagram.NodePadding = 5;
                SankeyDiagram.LabelPadding = 4;
                SankeyDiagram.RoleNames = {
                    rows: "Source",
                    columns: "Destination",
                    values: "Weight"
                };
                SankeyDiagram.capabilities = {
                    dataRoles: [{
                        name: SankeyDiagram.RoleNames.rows,
                        kind: powerbi.VisualDataRoleKind.Grouping,
                        displayName: SankeyDiagram.RoleNames.rows
                    }, {
                        name: SankeyDiagram.RoleNames.columns,
                        kind: powerbi.VisualDataRoleKind.Grouping,
                        displayName: SankeyDiagram.RoleNames.columns
                    }, {
                        name: SankeyDiagram.RoleNames.values,
                        kind: powerbi.VisualDataRoleKind.Measure,
                        displayName: SankeyDiagram.RoleNames.values
                    }],
                    dataViewMappings: [{
                        conditions: [
                            { "Source": { min: 0, max: 1 }, "Destination": { min: 0, max: 1 }, "Weight": { min: 0, max: 0 } },
                            { "Source": { min: 0, max: 1 }, "Destination": { min: 0, max: 1 }, "Weight": { min: 1, max: 1 } }
                        ],
                        categorical: {
                            categories: {
                                for: { in: SankeyDiagram.RoleNames.rows },
                                dataReductionAlgorithm: { top: {} }
                            },
                            values: {
                                select: [
                                    { bind: { to: SankeyDiagram.RoleNames.columns } },
                                    { bind: { to: SankeyDiagram.RoleNames.values } }
                                ]
                            }
                        }
                    }],
                    objects: {
                        general: {
                            displayName: powerbi.data.createDisplayNameGetter("Visual_General"),
                            properties: {
                                formatString: { type: { formatting: { formatString: true } } }
                            }
                        },
                        labels: {
                            displayName: powerbi.data.createDisplayNameGetter('Visual_DataPointsLabels'),
                            properties: {
                                show: {
                                    displayName: powerbi.data.createDisplayNameGetter("Visual_Show"),
                                    type: { bool: true }
                                },
                                fill: {
                                    displayName: powerbi.data.createDisplayNameGetter('Visual_Fill'),
                                    type: { fill: { solid: { color: true } } }
                                },
                                fontSize: {
                                    displayName: powerbi.data.createDisplayNameGetter('Visual_TextSize'),
                                    type: { formatting: { fontSize: true } }
                                }
                            }
                        },
                        links: {
                            displayName: "Links",
                            properties: {
                                fill: {
                                    displayName: powerbi.data.createDisplayNameGetter('Visual_Fill'),
                                    type: { fill: { solid: { color: true } } }
                                }
                            }
                        }
                    }
                };
                SankeyDiagram.Properties = SankeyDiagram.getProperties(SankeyDiagram.capabilities);
                return SankeyDiagram;
            })();
            SankeyDiagram1446463184954.SankeyDiagram = SankeyDiagram;
        })(SankeyDiagram1446463184954 = visuals.SankeyDiagram1446463184954 || (visuals.SankeyDiagram1446463184954 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var plugins;
        (function (plugins) {
            plugins.SankeyDiagram1446463184954 = {
                name: 'SankeyDiagram1446463184954',
                class: 'SankeyDiagram1446463184954',
                capabilities: powerbi.visuals.SankeyDiagram1446463184954.SankeyDiagram.capabilities,
                custom: true,
                create: function () { return new powerbi.visuals.SankeyDiagram1446463184954.SankeyDiagram(); }
            };
        })(plugins = visuals.plugins || (visuals.plugins = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
